/*
 *	httpsFwd.c
 *	Release $Name: MATRIXSSL_1_8_OPEN $
 *
 *	Simple example program for MatrixSSL
 *	Accepts a HTTPS request and echos the response back to the sender.
 */
/*
 *	Copyright (c) PeerSec Networks, 2002-2006. All Rights Reserved.
 *	The latest version of this code is available at http://www.matrixssl.org
 *
 *	This software is open source; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This General Public License does NOT permit incorporating this software 
 *	into proprietary programs.  If you are unable to comply with the GPL, a 
 *	commercial license for this software may be purchased from PeerSec Networks
 *	at http://www.peersec.com
 *	
 *	This program is distributed in WITHOUT ANY WARRANTY; without even the 
 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *	See the GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	http://www.gnu.org/copyleft/gpl.html
 */
/******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/wait.h>
#include <getopt.h>
#include <netdb.h>

/******************************************************************************/

#include <matrixSsl.h>
#include "sslSocket.h"

extern char *optarg;
extern int optind, opterr, optopt;

static int certChecker(sslCertInfo_t *cert, void *arg);

void sigchld_handler(int s)
{
	while(waitpid(-1, NULL, WNOHANG) > 0);
}

void usage()
{
	fprintf(stderr,"usage: FIXME\n");
}

/******************************************************************************/
/*
	This example application acts as an https server that accepts incoming
	client requests and reflects incoming data back to that client.  
*/
#if VXWORKS
int _httpsFwd(char *arg1)
#elif WINCE
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
					LPWSTR lpCmdLine, int nCmdShow)
#else
int main(int argc, char **argv)
#endif
{
	sslConn_t		*cp = NULL;
	sslKeys_t		*keys;
	SOCKET			srv_fd, src_fd, dst_fd;
	WSADATA			wsaData;
	int				status, src_rc, dst_rc;
	sslSessionId_t	*sessionId = NULL; // client mode
	short				cipherSuite = 0;
	unsigned char	buf[1024];
	void*				cervalidator = NULL;

	// options
	int				isClient = 0;
	char				*srv_host = NULL;
	char				*dst_host = "localhost";
	int				srv_port = 0;
	int				dst_port	= 0;
	char 				*keyfile = NULL; //"privkeySrv.pem";
	char 				*certfile = NULL; //"certSrv.pem";
	int				vlevel = 0;
	char 				*cpos,*opos;
	int tmpport;
				

	int c;
 	
#if VXWORKS
	int					argc;
	char				**argv;
	parseCmdLineArgs(arg1, &argc, &argv);
#endif /* VXWORKS */

#if WINCE
	int					argc;
	char				**argv;
	char				args[256];

/*
 *	parseCmdLineArgs expects an ASCII string and CE is unicoded, so convert
 *	the command line.  args will get hacked up, so you can't pass in a
 *	static string.
 */
	WideCharToMultiByte(CP_ACP, 0, lpCmdLine, -1, args, 256, NULL, NULL);

/*
 *	Parse the command line into an argv array.  This allocs memory, so
 *	we have to free argv when we're done.
 */
	parseCmdLineArgs(args, &argc, &argv);
#endif /* WINCE */

/*
	getopt
*/

	for (;;) {
		c = getopt (argc, argv, "cd:r:p:a:v:h");
		if (c == -1) {
			break;
		}

		switch (c) {
			case 'c':
				// client mode
				isClient=1;
				break;
			
			case 'd':
				// daemon mode [host:]port
				cpos = NULL;
				tmpport = 0;
				if((cpos = strchr(optarg,':'))) {				
					*cpos = '\0';
					if(optarg && optarg[0])
						srv_host = optarg;
					optarg = ++cpos;
				}
				if(optarg && optarg[0]) {
					tmpport = (int)strtol(optarg, (char **)NULL, 0);
					if(tmpport) srv_port = tmpport;
				}
				break;
			
			case 'r':
				// remote [host:]port
				cpos = NULL;
				tmpport = 0;
				if((cpos = strchr(optarg,':'))) {				
					*cpos = '\0';
					if(optarg && optarg[0])
						dst_host = optarg;
					optarg = ++cpos;
				}
				if(optarg && optarg[0]) {
					tmpport = (int)strtol(optarg, (char **)NULL, 0);
					if(tmpport) dst_port = tmpport;
				}
				break;
				
			case 'p':
				// pemfile (requred in servermode)
				keyfile = optarg;
				break;
			
			case 'a':
				// CA file
				certfile = optarg;
				break;
			
			case 'v':
				// veryfication level
				if(optarg && optarg[0]) {
					vlevel = (int)strtol(optarg, (char **)NULL, 0);
					if(vlevel == 1 ) {
						cervalidator = certChecker;
					}
					else if(vlevel > 3 || vlevel < 0) {
						fprintf(stderr,"-v takes whole numbers between 0 and 3");
						exit(2);
					}
				}
				break;
				
			case '?':
				usage();
				exit(2);
				break;
				
			case 'h':
				usage();
				exit(0);
				break;
		}
	}


/*
	Initialize Windows sockets (no-op on other platforms)
*/
	WSAStartup(MAKEWORD(1,1), &wsaData);
	
/*
	Initialize the MatrixSSL Library, and read in the public key (certificate)
	and private key.
*/
	if (matrixSslOpen() < 0) {
		fprintf(stderr, "matrixSslOpen failed, exiting...");
		exit(1);
	}

/*
	Standard PEM files
*/
	if (matrixSslReadKeys(&keys, certfile, keyfile, NULL, NULL) < 0)  {
		fprintf(stderr, "Error reading or parsing %s or %s.\n", 
			certfile, keyfile);
		exit(1);
	}

/*
	Create the listen socket
*/
	if ((srv_fd = socketListen(srv_port, &status)) == INVALID_SOCKET) {
		fprintf(stderr, "Cannot listen on port %d\n", srv_port);
		exit(1);
	}
/*
	Set blocking or not on the listen socket
*/
	setSocketBlock(srv_fd);
/*
	Loop control initalization
*/

	struct sigaction sa;
	sa.sa_handler = sigchld_handler; // reap all dead processes
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
		perror("sigaction");
		exit(1);
	}

/*
	Main connection loop
*/
	for (;;) {
/*
		sslAccept creates a new server session
*/
		setSocketBlock(srv_fd);
		if ((src_fd = socketAccept(srv_fd, &status)) == INVALID_SOCKET) {
			fprintf(stderr, "Error accepting connection: %d\n", status);
			continue;
		}		
		fprintf(stderr, "Accepting new connection.\n");				
		
		if (!fork()) {
			// this is the child process
			// socketShutdown(srv_fd); // child doesn't need the listener
			
			if (!isClient) {
				fprintf(stderr, "Trying to establish ssl connection ... ");
				if (sslAccept(&cp, src_fd, keys, cervalidator, 0)) {
					fprintf(stderr, "\nError could not establish ssl connection\n");
					socketShutdown(src_fd);
					exit(1);
				}
				fprintf(stderr,"ok\n");
			}
			
			/* try to connect to remote end of tunnel */
			fprintf(stderr, "Trying to connect server %s:%d ... ", dst_host, dst_port);
			if ((dst_fd = socketConnect(dst_host, dst_port, &status)) == INVALID_SOCKET) {
				fprintf(stderr, "\nError connecting to server %s:%d\n", dst_host, dst_port);
				exit(1);
			}
			fprintf(stderr, "ok\n");
			
			if (isClient) {
				fprintf(stderr, "Trying to establish an ssl connection to server %s:%d ... ", dst_host, dst_port);
				if ((dst_rc = sslConnect(&cp, dst_fd, keys, &sessionId, cipherSuite, cervalidator)) == INVALID_SOCKET) {
					fprintf(stderr, "\nError connecting to ssl server %s:%d\n", dst_host, dst_port);
					exit(1);
				}
				fprintf(stderr, "ok\n");
			}

			
			/* purge read buffer to prevent blocking on linux*/
			
			fd_set rfds;
			SOCKET plain_fd = isClient ? cp->fd : dst_fd;
			SOCKET max = cp->fd > dst_fd ? cp->fd : dst_fd;
			
			fprintf(stderr, "Starting Proxy %d <-> %s:%d\n", srv_port, dst_host, dst_port);
			setSocketNonblock(plain_fd);
			setSocketNonblock(cp->fd);
						
			for(;;) {
				struct timeval tv = {300,0};
				FD_ZERO(&rfds);
				FD_SET(plain_fd, &rfds);
				FD_SET(cp->fd, &rfds);
				
				fprintf(stderr, "DEBUG. select begin\n");
				select(max+1, &rfds, NULL, NULL, &tv);
				fprintf(stderr, "DEBUG. select end\n");				
				

				fprintf(stderr, "DEBUG. read ssl - write plain... ");
				src_rc=0;
				if (FD_ISSET(cp->fd,&rfds) && (src_rc = sslRead(cp, buf, sizeof(buf)-1, &status)) > 0) {
					/* forward to tunnel endpoint */
					send(plain_fd, buf, src_rc, MSG_NOSIGNAL);
				}
				fprintf(stderr, "%d bytes\n",src_rc);
							
				// read response from tunnel endpoint
				fprintf(stderr, "DEBUG. read palin - write ssl... ");
				dst_rc=0;
				if (FD_ISSET(plain_fd,&rfds) && (dst_rc = recv(plain_fd, buf, sizeof(buf)-1, MSG_NOSIGNAL)) > 0) {
					/* write back to https client */
					sslWrite(cp, buf, dst_rc, &status);
				}
				fprintf(stderr, "%d bytes\n",dst_rc);
				
				if (src_rc < 0 || dst_rc < 0) {
					fprintf(stderr, "connection error. closing connection.\n");
					break;
				}
				else if (src_rc == 0 && dst_rc == 0) {
					fprintf(stderr, "done. closing connection.\n");
					break;				
				}
			}
			

			sslWriteClosureAlert(cp);
			socketShutdown(cp->fd);
			sslFreeConnection(&cp);
	
			socketShutdown(plain_fd);
			exit(0);
		}
		//socketShutdown(src_fd);  // parent doesn't need this
	}
/*
	Close listening socket, free remaining items
*/
	if (cp && cp->ssl) {
		socketShutdown(cp->fd);
		sslFreeConnection(&cp);
	}
	socketShutdown(srv_fd);

	matrixSslFreeKeys(keys);
	matrixSslClose();
	WSACleanup();
	return 0;
}

/******************************************************************************/
/*
	Stub for a user-level certificate validator.  Just using
	the default validation value here.
*/
static int certChecker(sslCertInfo_t *cert, void *arg)
{
	sslCertInfo_t	*next;
/*
	Make sure we are checking the last cert in the chain
*/
	next = cert;
	while (next->next != NULL) {
		next = next->next;
	}
#if ENFORCE_CERT_VALIDATION
/*
	This case passes the true RSA authentication status through
*/
	return next->verified;
#else
/*
	This case passes an authenticated server through, but flags a
	non-authenticated server correctly.  The user can call the
	matrixSslGetAnonStatus later to see the status of this connection.
*/
	if (next->verified != 1) {
		return 1;//SSL_ALLOW_ANON_CONNECTION;
	}
	return next->verified;
#endif /* ENFORCE_CERT_VALIDATION */
}		

/******************************************************************************/


